package cn.zgfxrc.boot.upms.biz.service;

import cn.zgfxrc.boot.common.core.entity.base.BaseModel;
import cn.zgfxrc.boot.common.core.entity.base.DragModel;
import cn.zgfxrc.boot.common.core.enums.ActiveStatus;
import cn.zgfxrc.boot.common.core.enums.DropPosition;
import cn.zgfxrc.boot.common.core.util.TenantUtil;
import cn.zgfxrc.boot.common.core.util.TreeUtil;
import cn.zgfxrc.boot.upms.api.entity.SysDept;
import cn.zgfxrc.boot.upms.api.entity.SysDeptUser;
import cn.zgfxrc.boot.upms.api.entity.SysUser;
import cn.zgfxrc.boot.upms.api.entity.res.SysDeptRes;
import cn.zgfxrc.boot.upms.api.entity.res.SysDeptTreeNode;
import cn.zgfxrc.boot.upms.api.enums.DeptUserType;
import cn.zgfxrc.boot.upms.biz.mapper.SysDeptMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysDeptUserMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysUserMapper;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.conditions.update.LambdaUpdateChainWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.seepine.secret.AuthUtil;
import com.seepine.secret.entity.AuthUser;
import com.seepine.tool.Run;
import com.seepine.tool.cache.Cache;
import com.seepine.tool.lock.Lock;
import com.seepine.tool.secure.digest.MD5;
import com.seepine.tool.util.ListUtil;
import com.seepine.tool.util.Objects;
import com.seepine.tool.util.Strings;
import com.seepine.tool.util.Validate;
import java.io.Serializable;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import lombok.AllArgsConstructor;
import org.springframework.beans.BeanUtils;
import org.springframework.lang.Nullable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author huanghs
 * @date 2022.12.2
 */
@Service
@AllArgsConstructor
@Transactional(rollbackFor = Exception.class)
public class SysDeptService extends ServiceImpl<SysDeptMapper, SysDept> {
  private final SysDeptUserMapper sysDeptUserMapper;
  private final SysUserMapper sysUserMapper;

  @Override
  public boolean removeById(Serializable id) {
    Validate.isTrue(
        sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getDeptId, id)
                .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                .count()
            == 0,
        "部门中存在成员，请先移除后再删除部门");
    return super.removeById(id);
  }

  @Override
  public boolean save(SysDept entity) {
    return Lock.sync(
        "upms-sys-dept-save",
        () -> {
          entity.setName(Validate.nonBlank(entity.getName(), "部门名称不能为空"));
          Validate.isTrue(
              super.lambdaQuery().eq(SysDept::getName, entity.getName()).count() == 0, "部门名称已存在");
          entity.setParentId(entity.getParentId() == null ? 0L : entity.getParentId());
          entity.setSort(
              Optional.ofNullable(
                          super.lambdaQuery()
                              .eq(SysDept::getParentId, entity.getParentId())
                              .orderByDesc(SysDept::getSort)
                              .last("limit 1")
                              .one())
                      .orElseGet(
                          () -> {
                            SysDept dept = new SysDept();
                            dept.setSort(-1);
                            return dept;
                          })
                      .getSort()
                  + 1);
          return super.save(entity);
        });
  }

  /**
   * 获取部门树结构包含租户根
   *
   * @param entity req
   * @param draggable 是否可拖拽
   * @return treeList
   */
  public List<SysDeptTreeNode> treeListWithTenant(
      @Nullable SysDept entity, boolean draggable, boolean withChargeUserIds) {
    SysDeptTreeNode root = new SysDeptTreeNode();
    AuthUser user = AuthUtil.getUser();
    root.setName(user.getTenantName());
    root.setId(0L);
    root.setParentId(-1L);
    root.setDraggable(false);
    root.setChildren(treeList(entity, draggable, withChargeUserIds));
    return Collections.singletonList(root);
  }
  /**
   * 获取部门树结构
   *
   * @param entity req
   * @param draggable 是否可拖拽
   * @return treeList
   */
  public List<SysDeptTreeNode> treeList(
      @Nullable SysDept entity, boolean draggable, boolean withChargeUserIds) {
    LambdaQueryWrapper<SysDept> wrapper = Wrappers.lambdaQuery();
    if (entity != null) {
      Run.nonBlankAndTrim(entity.getName(), val -> wrapper.like(SysDept::getName, val));
    }
    return TreeUtil.build(
        list(wrapper).stream()
            .map(
                (item) -> {
                  SysDeptTreeNode node = new SysDeptTreeNode();
                  BeanUtils.copyProperties(item, node);
                  node.setDraggable(draggable);
                  if (withChargeUserIds) {
                    var primaryIds =
                        ListUtil.map(
                            sysDeptUserMapper
                                .lambdaQuery()
                                .eq(SysDeptUser::getDeptId, item.getId())
                                .eq(SysDeptUser::getType, DeptUserType.PRIMARY)
                                .list(),
                            SysDeptUser::getUserId);
                    node.setPrimaryChargeUserId(primaryIds.isEmpty() ? null : primaryIds.get(0));
                    Run.nonNull(
                        node.getPrimaryChargeUserId(),
                        id ->
                            node.setPrimaryChargeUserName(
                                Optional.ofNullable(
                                        sysUserMapper.lambdaQuery().eq(BaseModel::getId, id).one())
                                    .map(SysUser::getFullName)
                                    .orElse("")));
                    node.setSecondaryChargeUserIds(
                        ListUtil.map(
                            sysDeptUserMapper
                                .lambdaQuery()
                                .eq(SysDeptUser::getDeptId, item.getId())
                                .eq(SysDeptUser::getType, DeptUserType.SECONDARY)
                                .list(),
                            SysDeptUser::getUserId));
                    Run.nonEmpty(
                        node.getSecondaryChargeUserIds(),
                        ids ->
                            node.setSecondaryChargeUserNames(
                                Optional.ofNullable(
                                        sysUserMapper
                                            .lambdaQuery()
                                            .in(BaseModel::getId, ids)
                                            .list())
                                    .orElse(List.of())
                                    .stream()
                                    .map(SysUser::getFullName)
                                    .collect(Collectors.joining(Strings.COMMA))));
                  }
                  return node;
                })
            .toList());
  }
  /**
   * 获取部门树结构包含用户
   *
   * @return treeList
   */
  public List<SysDeptTreeNode> fetchUser() {
    LambdaQueryWrapper<SysDept> wrapper = Wrappers.lambdaQuery();
    List<SysDept> deptList = new ArrayList<>();
    SysDept root = new SysDept();
    AuthUser authUser = AuthUtil.getUser();
    root.setName(authUser.getTenantName());
    root.setId(0L);
    root.setParentId(-1L);
    root.setSort(0);
    deptList.add(root);
    deptList.addAll(list(wrapper.orderByAsc(SysDept::getSort)));

    List<SysDeptTreeNode> deptNodeList =
        deptList.stream()
            .flatMap(
                (item) -> {
                  SysDeptTreeNode node = new SysDeptTreeNode();
                  BeanUtils.copyProperties(item, node);
                  node.setDraggable(false);
                  node.setType("dept");
                  node.setCheckable(false);
                  // 根据部门id查询用户ids
                  List<Long> userIds =
                      ListUtil.map(
                          sysDeptUserMapper
                              .lambdaQuery()
                              .eq(SysDeptUser::getDeptId, item.getId())
                              .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                              .list(),
                          SysDeptUser::getUserId);
                  if (!userIds.isEmpty()) {
                    var userList =
                        sysUserMapper
                            .lambdaQuery()
                            .in(BaseModel::getId, userIds)
                            // 已停用的账号不展示在部门下
                            .ne(SysUser::getStatus, ActiveStatus.SUSPEND)
                            .list();
                    AtomicInteger i = new AtomicInteger(-userList.size() - 1);
                    var userNodeSteam =
                        userList.stream()
                            .map(
                                user -> {
                                  SysDeptTreeNode userNode = new SysDeptTreeNode();
                                  userNode.setDraggable(false);
                                  userNode.setType("user");
                                  userNode.setId(user.getId());
                                  userNode.setName(user.getFullName());
                                  userNode.setAvatarUrl(user.getAvatarUrl());
                                  userNode.setParentId(item.getId());
                                  userNode.setCheckable(true);
                                  userNode.setSort(node.getSort() + i.getAndIncrement());
                                  return userNode;
                                });
                    return Stream.concat(Stream.of(node), userNodeSteam);
                  }
                  return Stream.of(node);
                })
            .toList();
    return TreeUtil.build(deptNodeList, -1L);
  }
  /**
   * 部门拖拽排序
   *
   * @param req req
   */
  public void drag(DragModel req) {
    Lock.sync(
        "sys_dept_drag",
        () -> {
          SysDept drag = Validate.nonNull(super.getById(req.getDragId()), "找不到拖拽部门数据，请刷新页面后重试");
          SysDept drop = Validate.nonNull(super.getById(req.getDropId()), "找不到拖拽后部门数据，请刷新页面后重试");
          LambdaUpdateChainWrapper<SysDept> dragWrapper =
              lambdaUpdate().eq(BaseModel::getId, drag.getId());
          LambdaUpdateChainWrapper<SysDept> dropWrapper = lambdaUpdate();
          // 1.移入里面
          if (req.getDropPosition().equals(DropPosition.INNER)) {
            // 拖拽对象parentId修改，并sort设为0
            dragWrapper.set(SysDept::getParentId, drop.getId()).set(SysDept::getSort, 0);
            // 释放对象下的所有sort+1
            dropWrapper.eq(SysDept::getParentId, drop.getId()).setSql("sort=sort+1");
          }
          // 2.移到上方
          else if (req.getDropPosition().equals(DropPosition.UPPER)) {
            // 拖拽对象parentId修改，并sort设为释放对象sort+1
            dragWrapper
                .set(SysDept::getParentId, drop.getParentId())
                .set(SysDept::getSort, drop.getSort());
            // 释放对象sort及以上的所有sort+1
            dropWrapper
                .eq(SysDept::getParentId, drop.getParentId())
                .ge(SysDept::getSort, drop.getSort())
                .setSql("sort=sort+1");
          }
          // 3.移到下方
          else {
            // 拖拽对象parentId修改，并sort设为释放对象sort+1
            dragWrapper
                .set(SysDept::getParentId, drop.getParentId())
                .set(SysDept::getSort, drop.getSort() + 1);
            // 释放对象sort及以上的所有sort+1
            dropWrapper
                .eq(SysDept::getParentId, drop.getParentId())
                .gt(SysDept::getSort, drop.getSort())
                .setSql("sort=sort+1");
          }
          // 先修改其他影响sort
          dropWrapper.update();
          // 再修改拖拽对象
          dragWrapper.update();
        });
  }

  /**
   * 重置用户部门关联
   *
   * @param userId 用户id
   * @param deptIds 新部门ids
   * @date 2022.12.6
   */
  @Transactional(rollbackFor = Exception.class)
  public void resetUserDept(Long userId, List<Long> deptIds) {
    Validate.nonEmpty(deptIds, "部门不能为空");
    // 1. 获取当前所属部门ids
    List<Long> nowDeptIds =
        ListUtil.map(
            sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                .eq(SysDeptUser::getUserId, userId)
                .list(),
            SysDeptUser::getDeptId);
    // 2. 删除当前所有部门
    Run.nonEmpty(
        nowDeptIds,
        delIds ->
            // 2.1 删除关联关系
            Validate.isTrue(
                sysDeptUserMapper.delete(
                        Wrappers.<SysDeptUser>lambdaQuery()
                            .eq(SysDeptUser::getUserId, userId)
                            .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                            .in(SysDeptUser::getDeptId, delIds))
                    > 0,
                "移除旧部门数据失败"));
    // 3. 生成新的
    Validate.isTrue(
        sysDeptUserMapper.insertBatchSomeColumn(
                ListUtil.map(
                    deptIds,
                    deptId -> {
                      SysDeptUser deptUser = new SysDeptUser();
                      deptUser.setUserId(userId);
                      deptUser.setDeptId(deptId);
                      deptUser.setType(DeptUserType.NORMAL);
                      deptUser.setTenantId(Long.valueOf(TenantUtil.getTenantId()));
                      return deptUser;
                    }))
            > 0,
        "保存失败");
  }

  /**
   * 获取部门ids
   *
   * @param userId 用户id
   * @return 部门ids
   */
  public List<SysDeptUser> getDeptUsers(Long userId) {
    List<SysDeptUser> list =
        sysDeptUserMapper
            .lambdaQuery()
            .eq(SysDeptUser::getType, DeptUserType.NORMAL)
            .eq(SysDeptUser::getUserId, userId)
            .orderByAsc(BaseModel::getId)
            .list();
    return Objects.isEmpty(list) ? new ArrayList<>() : list;
  }

  /**
   * 获取部门list，会包含租户
   *
   * @param ids ids
   * @return deptList
   */
  public List<SysDept> getListByIds(List<Long> ids) {
    AuthUser authUser = AuthUtil.getUser();
    List<SysDept> list = listByIds(ids);
    if (ids.contains(0L)) {
      SysDept dept = new SysDept();
      dept.setId(0L);
      dept.setName(authUser.getTenantName());
      dept.setParentId(-1L);
      list.add(0, dept);
    }
    return list;
  }

  /**
   * 通过父id获取包含父id的所有ids
   *
   * @param parentId 父id
   * @return 包含父id的所有ids
   */
  public List<Long> findAllChildrenIds(Long parentId) {
    Set<Long> ids = new HashSet<>();
    ids.add(parentId);
    if (parentId.equals(0L)) {
      ids.addAll(ListUtil.map(super.lambdaQuery().list(), BaseModel::getId));
    } else {
      List<Long> childrenIds =
          ListUtil.map(
              super.lambdaQuery().eq(SysDept::getParentId, parentId).list(), BaseModel::getId);
      for (Long childId : childrenIds) {
        ids.addAll(findAllChildrenIds(childId));
      }
    }
    return new ArrayList<>(ids);
  }

  /**
   * 通过部门id获取用户ids
   *
   * @param deptId 部门id
   * @return 用户ids
   * @date 2022.12.12
   */
  public List<Long> userIdsByDeptId(Long deptId) {
    return ListUtil.map(
        sysDeptUserMapper
            .lambdaQuery()
            .eq(SysDeptUser::getDeptId, deptId)
            .eq(SysDeptUser::getType, DeptUserType.NORMAL)
            .list(),
        SysDeptUser::getUserId);
  }

  /**
   * 重置部门与成员关联
   *
   * @param deptId 用户id
   * @param userIds 新成员ids
   * @param userType 用户类型(普通成员或负责人)
   * @date 2022.12.12
   */
  @Transactional(rollbackFor = Exception.class)
  public void resetDeptUser(Long deptId, List<Long> userIds, DeptUserType userType) {
    Validate.nonEmpty(deptId, "部门id不能为空");
    Validate.nonEmpty(userType, "用户类型不能为空");
    if (userIds == null) {
      userIds = new ArrayList<>();
    }
    userIds = userIds.stream().filter(java.util.Objects::nonNull).toList();
    // 1. 获取当前部门对应用户类型的用户ids
    List<Long> nowUserIds =
        ListUtil.map(
            sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getType, userType)
                .eq(SysDeptUser::getDeptId, deptId)
                .list(),
            SysDeptUser::getUserId);
    // 2. 过滤出需要删除的,now=[1,2,3],new=[3,4],needDel=[1,2]
    List<Long> finalUserIds = userIds;
    Run.nonEmpty(
        ListUtil.filter(nowUserIds, userId -> !finalUserIds.contains(userId)),
        delIds -> {
          // 2.1 类型如果是普通成员，需要判断
          if (userType.equals(DeptUserType.NORMAL)) {
            // 2.1.1 查询这些用户，非当前部门的关联数据
            Set<Long> otherUserIds =
                new HashSet<>(
                    ListUtil.map(
                        sysDeptUserMapper
                            .lambdaQuery()
                            .select(SysDeptUser::getUserId)
                            .eq(SysDeptUser::getType, userType)
                            .ne(SysDeptUser::getDeptId, deptId)
                            .in(SysDeptUser::getUserId, delIds)
                            .list(),
                        SysDeptUser::getUserId));
            // 2.1.2 如果某个用户不在里面了，不允许删除，因为用户至少需要存在一个部门
            if (otherUserIds.size() != new HashSet<>(delIds).size()) {
              // 找出不存在了的用户ids
              List<Long> onlyIds = ListUtil.filter(delIds, id -> !otherUserIds.contains(id));
              if (!onlyIds.isEmpty()) {
                String fullNames =
                    sysUserMapper
                        .lambdaQuery()
                        .select(SysUser::getFullName)
                        .in(BaseModel::getId, onlyIds)
                        .list()
                        .stream()
                        .map(SysUser::getFullName)
                        .collect(Collectors.joining(","));
                // 必须为blank，不为blank表示找到仅存用户，需抛异常
                Validate.isBlank(fullNames, "成员：[" + fullNames + "] 仅归属于当前部门，无法移除");
              }
            }
          }
          // 2.2 删除关联关系
          Validate.isTrue(
              sysDeptUserMapper.delete(
                      Wrappers.<SysDeptUser>lambdaQuery()
                          .eq(SysDeptUser::getType, userType)
                          .eq(SysDeptUser::getDeptId, deptId)
                          .in(SysDeptUser::getUserId, delIds))
                  > 0,
              "移除旧成员数据失败");
        });
    // 3. 过滤出已经存在的不需要重新插入,now=[1,2,3],new=[3,4],needAdd=[4]
    List<Long> needAdUserIds = ListUtil.filter(userIds, userId -> !nowUserIds.contains(userId));
    if (needAdUserIds.isEmpty()) {
      return;
    }
    if (userType.equals(DeptUserType.PRIMARY)) {
      sysDeptUserMapper.delete(
          Wrappers.<SysDeptUser>lambdaQuery()
              .eq(SysDeptUser::getType, DeptUserType.PRIMARY)
              .eq(SysDeptUser::getDeptId, deptId));
    }
    // 4. 生成新的
    AuthUser authUser = AuthUtil.getUser();
    LocalDateTime nowTime = LocalDateTime.now();
    Validate.isTrue(
        sysDeptUserMapper.insertBatchSomeColumn(
                ListUtil.map(
                    needAdUserIds,
                    userId -> {
                      SysDeptUser deptUser = new SysDeptUser();
                      deptUser.setUserId(userId);
                      deptUser.setDeptId(deptId);
                      deptUser.setCreateId(authUser.getIdAsLong());
                      deptUser.setCreateName(authUser.getFullName());
                      deptUser.setCreateTime(nowTime);
                      deptUser.setType(userType);
                      deptUser.setTenantId(authUser.getTenantIdAsLong());
                      return deptUser;
                    }))
            > 0,
        "保存失败");
  }

  /**
   * 移除部门中某个成员
   *
   * @param req {deptId,userId}
   * @date 2022.12.21
   */
  @Transactional(rollbackFor = Exception.class)
  public void removeDeptUser(SysDeptUser req) {
    Validate.nonEmpty(req.getDeptId(), "部门参数不能为空");
    Validate.nonEmpty(req.getUserId(), "用户参数不能为空");
    Lock.sync(
        "remove_dept_user_" + req.getDeptId() + "_" + req.getUserId(),
        () -> {
          Validate.isTrue(
              sysDeptUserMapper.delete(
                      Wrappers.<SysDeptUser>lambdaQuery()
                          .eq(SysDeptUser::getDeptId, req.getDeptId())
                          .eq(SysDeptUser::getUserId, req.getUserId())
                          .eq(SysDeptUser::getType, DeptUserType.NORMAL))
                  >= 0,
              "删除失败");
          Validate.isTrue(
              sysDeptUserMapper
                      .lambdaQuery()
                      .eq(SysDeptUser::getUserId, req.getUserId())
                      .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                      .count()
                  > 0,
              "该成员仅归属于当前部门，无法移除");
        });
  }

  /**
   * 删除部门
   *
   * @param dept 部门
   */
  public void del(SysDept dept) {
    Validate.nonNull(dept.getId(), "部门id不能为空");
    Validate.isTrue(
        super.lambdaQuery().eq(SysDept::getParentId, dept.getId()).count() <= 0, "删除失败，请先移除子部门");
    Validate.isTrue(
        sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getDeptId, dept.getId())
                .eq(SysDeptUser::getType, DeptUserType.NORMAL)
                .count()
            <= 0,
        "删除失败，请先移除部门下员工");
    Validate.isTrue(super.removeById(dept.getId()), "删除失败，请重试");
  }

  public SysDeptRes getResById(Long id) {
    Validate.nonNull(id, "参数错误");
    SysDept find = Validate.nonNull(super.getById(id), "部门不存在");
    SysDeptRes res = new SysDeptRes();
    BeanUtils.copyProperties(find, res);
    // 设置主负责人
    var primaryIds =
        ListUtil.map(
            sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getDeptId, find.getId())
                .eq(SysDeptUser::getType, DeptUserType.PRIMARY)
                .list(),
            SysDeptUser::getUserId);
    res.setPrimaryChargeUserId(primaryIds.isEmpty() ? null : primaryIds.get(0));
    Run.nonNull(
        res.getPrimaryChargeUserId(),
        val ->
            res.setPrimaryChargeUserName(
                Optional.ofNullable(sysUserMapper.lambdaQuery().eq(BaseModel::getId, val).one())
                    .map(SysUser::getFullName)
                    .orElse("")));
    // 设置副负责人
    res.setSecondaryChargeUserIds(
        ListUtil.map(
            sysDeptUserMapper
                .lambdaQuery()
                .eq(SysDeptUser::getDeptId, id)
                .eq(SysDeptUser::getType, DeptUserType.SECONDARY)
                .list(),
            SysDeptUser::getUserId));
    Run.nonEmpty(
        res.getSecondaryChargeUserIds(),
        ids ->
            res.setSecondaryChargeUserNames(
                Optional.ofNullable(
                        Cache.getIfPresent(
                            "sys_user:" + MD5.digestHex(ids.toString()),
                            () -> sysUserMapper.lambdaQuery().in(BaseModel::getId, ids).list(),
                            Duration.ofSeconds(30).toMillis()))
                    .orElse(List.of())
                    .stream()
                    .map(SysUser::getFullName)
                    .collect(Collectors.joining(Strings.COMMA))));
    return res;
  }
}
