package cn.zgfxrc.boot.upms.biz.service;

import cn.zgfxrc.boot.common.core.constant.Constant;
import cn.zgfxrc.boot.common.core.entity.base.BaseModel;
import cn.zgfxrc.boot.common.core.enums.PlatformType;
import cn.zgfxrc.boot.common.core.util.TenantUtil;
import cn.zgfxrc.boot.upms.api.entity.SysDict;
import cn.zgfxrc.boot.upms.api.entity.SysDictItem;
import cn.zgfxrc.boot.upms.api.enums.DictType;
import cn.zgfxrc.boot.upms.biz.mapper.SysDictMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysTenantMapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.seepine.secret.AuthUtil;
import com.seepine.tool.lock.Lock;
import com.seepine.tool.util.Validate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;
import lombok.AllArgsConstructor;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author huanghs
 * @date 2022.12.2
 */
@Service
@AllArgsConstructor
@Transactional(rollbackFor = Exception.class)
public class SysDictService extends ServiceImpl<SysDictMapper, SysDict> {
  private final SysDictItemService sysDictItemService;
  private final SysTenantMapper sysTenantMapper;

  @Override
  public boolean save(SysDict entity) {
    return Lock.sync(
        Validate.nonBlank(entity.getDicCode(), "字典编码不能为空"),
        () -> {
          Validate.isTrue(
              lambdaQuery().eq(SysDict::getDicCode, entity.getDicCode()).count() == 0,
              "字典编码[" + entity.getDicCode() + "]已存在");
          entity.setSort(
              Optional.ofNullable(
                          super.lambdaQuery()
                              .select(SysDict::getSort)
                              .orderByDesc(SysDict::getSort)
                              .last("limit 1")
                              .one())
                      .map(SysDict::getSort)
                      .orElse(-1)
                  + 1);
          // 保存默认字典时，为已存在租户插入值
          if (entity.getType().equals(DictType.DEFAULT)) {
            // 必须平台端才能保存默认字典项
            Validate.isTrue(
                PlatformType.SYS.eq(
                    PlatformType.from(AuthUtil.getUser().getClaim(Constant.platformType))),
                "非平台端无法创建默认字典");
            for (var sysTenant :
                sysTenantMapper
                    .lambdaQuery()
                    .lt(BaseModel::getCreateTime, LocalDateTime.now())
                    .ne(BaseModel::getId, TenantUtil.getTenantId())
                    .list()) {
              TenantUtil.run(
                  sysTenant.getId(),
                  () -> {
                    SysDict dict = new SysDict();
                    BeanUtils.copyProperties(entity, dict);
                    Validate.isTrue(super.save(dict), "保存失败");
                  });
            }
          }
          return super.save(entity);
        });
  }

  public List<SysDictItem> dictList(String code) {
    var superDict = TenantUtil.apply(0L, () -> lambdaQuery().eq(SysDict::getDicCode, code).one());
    Validate.nonNull(superDict, "字典编码错误");
    // 1. 全局字典直接返回
    if (superDict.getType().equals(DictType.GLOBAL)) {
      return TenantUtil.apply(
          superDict.getTenantId(), () -> sysDictItemService.list(superDict.getId()));
    }
    // 2. 默认字典找当前租户
    var currentDict = lambdaQuery().eq(SysDict::getDicCode, code).one();
    // 3. 若当前租户找到空,也返回平台的字典
    var dict = Objects.requireNonNullElse(currentDict, superDict);
    return TenantUtil.apply(dict.getTenantId(), () -> sysDictItemService.list(dict.getId()));
  }

  /**
   * 初始化新租户字典
   *
   * @param initTenantId 新租户id
   */
  @Transactional(rollbackFor = Exception.class)
  public void initDefaultDict(Long initTenantId) {
    Long supplierId = 0L;
    // 1.获取平台端所有默认字典
    var dictList =
        TenantUtil.apply(
            supplierId, () -> lambdaQuery().eq(SysDict::getType, DictType.DEFAULT).list());
    var dictItemMap = new HashMap<String, List<SysDictItem>>();
    dictList.parallelStream()
        .forEach(
            item ->
                dictItemMap.put(
                    item.getDicCode(),
                    TenantUtil.apply(supplierId, () -> sysDictItemService.list(item.getId()))));
    for (SysDict dict : dictList) {
      var entity = new SysDict();
      entity.setName(dict.getName());
      entity.setDicCode(dict.getDicCode());
      entity.setDescription(dict.getDescription());
      entity.setType(DictType.DEFAULT);
      entity.setSort(dict.getSort());
      // 保存字典
      Validate.isTrue(super.save(entity), "保存失败");
      // 保存字典项
      sysDictItemService
          .getBaseMapper()
          .insertBatchSomeColumn(
              dictItemMap.get(dict.getDicCode()).parallelStream()
                  .map(
                      item -> {
                        var dictItem = new SysDictItem();
                        dictItem.setDictId(entity.getId());
                        dictItem.setLabel(item.getLabel());
                        dictItem.setValue(item.getValue());
                        dictItem.setDescription(item.getDescription());
                        dictItem.setDisabled(item.getDisabled());
                        dictItem.setEditable(item.getEditable());
                        dictItem.setSort(item.getSort());
                        dictItem.setTenantId(initTenantId);
                        return dictItem;
                      })
                  .collect(Collectors.toList()));
    }
  }
}
