package cn.zgfxrc.boot.upms.biz.service;

import cn.zgfxrc.boot.upms.api.entity.*;
import cn.zgfxrc.boot.upms.api.entity.res.SysGroupRes;
import cn.zgfxrc.boot.upms.api.entity.res.SysUserRes;
import cn.zgfxrc.boot.upms.api.enums.DeptUserType;
import cn.zgfxrc.boot.upms.api.enums.GroupUserType;
import cn.zgfxrc.boot.upms.api.enums.GroupVisibleScope;
import cn.zgfxrc.boot.upms.biz.mapper.SysGroupMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysGroupUserMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysUserMapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.seepine.secret.AuthUtil;
import com.seepine.secret.entity.AuthUser;
import com.seepine.tool.Run;
import com.seepine.tool.lock.Lock;
import com.seepine.tool.util.ListUtil;
import com.seepine.tool.util.Strings;
import com.seepine.tool.util.Validate;
import java.io.Serializable;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;
import lombok.AllArgsConstructor;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author huanghs
 */
@Service
@AllArgsConstructor
@Transactional(rollbackFor = Exception.class)
public class SysGroupService extends ServiceImpl<SysGroupMapper, SysGroup> {
  private final SysGroupUserMapper sysGroupUserMapper;
  private final SysUserMapper sysUserMapper;

  public List<SysGroupRes> fetch(SysGroup entity) {
    return ListUtil.map(
        super.list(Wrappers.lambdaQuery(entity).orderByAsc(SysGroup::getSort)),
        item -> {
          SysGroupRes res = new SysGroupRes();
          BeanUtils.copyProperties(item, res);
          var primaryIds =
              ListUtil.map(sysGroupUserMapper.listPrimary(item.getId()), SysGroupUser::getUserId);
          res.setPrimaryChargeUserId(primaryIds.isEmpty() ? null : primaryIds.get(0));
          Run.nonNull(
              res.getPrimaryChargeUserId(),
              id ->
                  res.setPrimaryChargeUserName(
                      Optional.ofNullable(sysUserMapper.selectById(id))
                          .map(SysUser::getFullName)
                          .orElse("")));
          res.setSecondaryChargeUserIds(
              ListUtil.map(
                  sysGroupUserMapper.listSecondary(item.getId()), SysGroupUser::getUserId));
          Run.nonEmpty(
              res.getSecondaryChargeUserIds(),
              ids ->
                  res.setSecondaryChargeUserNames(
                      Optional.ofNullable(sysUserMapper.selectBatchIds(ids))
                          .orElse(List.of())
                          .stream()
                          .map(SysUser::getFullName)
                          .collect(Collectors.joining(Strings.COMMA))));
          return res;
        });
  }

  @Override
  public boolean removeById(Serializable id) {
    Validate.isTrue(
        sysGroupUserMapper
                .lambdaQuery()
                .eq(SysGroupUser::getGroupId, id)
                .eq(SysGroupUser::getType, GroupUserType.NORMAL)
                .count()
            == 0,
        "用户组中存在成员，请先移除后再删除用户组");
    return super.removeById(id);
  }

  @Override
  public boolean save(SysGroup entity) {
    return Lock.sync(
        "upms-sys-group-save",
        () -> {
          entity.setName(Validate.nonBlank(entity.getName(), "用户组名称不能为空"));
          Validate.isTrue(
              super.lambdaQuery().eq(SysGroup::getName, entity.getName()).count() == 0, "用户组名称已存在");
          entity.setSort(
              Optional.ofNullable(
                          super.lambdaQuery().orderByDesc(SysGroup::getSort).last("limit 1").one())
                      .orElseGet(
                          () -> {
                            SysGroup dept = new SysGroup();
                            dept.setSort(-1);
                            return dept;
                          })
                      .getSort()
                  + 1);
          entity.setVisibleScope(
              Objects.requireNonNullElse(entity.getVisibleScope(), GroupVisibleScope.PUBLIC));
          return super.save(entity);
        });
  }
  /**
   * 通过id获取用户ids
   *
   * @param groupId id
   * @return 用户ids
   * @date 2022.12.12
   */
  public List<Long> userIdsById(Long groupId) {
    return ListUtil.map(
        sysGroupUserMapper.list(groupId, GroupUserType.NORMAL), SysGroupUser::getUserId);
  }
  /**
   * 重置组与成员关联
   *
   * @param groupId 组id
   * @param userIds 新成员ids
   * @param groupType 组类型(普通成员或负责人)
   */
  @Transactional(rollbackFor = Exception.class)
  public void resetUser(Long groupId, List<Long> userIds, GroupUserType groupType) {
    Validate.nonEmpty(groupId, "组id不能为空");
    Validate.nonEmpty(groupType, "组类型不能为空");
    if (userIds == null) {
      userIds = new ArrayList<>();
    }
    userIds = userIds.stream().filter(java.util.Objects::nonNull).toList();
    // 1. 获取当前组对应用户类型的用户ids
    List<Long> nowUserIds =
        ListUtil.map(
            sysGroupUserMapper
                .lambdaQuery()
                .eq(SysGroupUser::getType, groupType)
                .eq(SysGroupUser::getGroupId, groupId)
                .list(),
            SysGroupUser::getUserId);
    // 2. 过滤出需要删除的,now=[1,2,3],new=[3,4],needDel=[1,2]
    List<Long> finalUserIds = userIds;
    Run.nonEmpty(
        ListUtil.filter(nowUserIds, userId -> !finalUserIds.contains(userId)),
        delIds -> {
          // 2.1 删除关联关系
          Validate.isTrue(
              sysGroupUserMapper.delete(
                      Wrappers.<SysGroupUser>lambdaQuery()
                          .eq(SysGroupUser::getType, groupType)
                          .eq(SysGroupUser::getGroupId, groupId)
                          .in(SysGroupUser::getUserId, delIds))
                  > 0,
              "移除旧成员数据失败");
        });
    // 3. 过滤出已经存在的不需要重新插入,now=[1,2,3],new=[3,4],needAdd=[4]
    List<Long> needAdUserIds = ListUtil.filter(userIds, userId -> !nowUserIds.contains(userId));
    if (needAdUserIds.isEmpty()) {
      return;
    }
    if (groupType.equals(GroupUserType.PRIMARY)) {
      sysGroupUserMapper.delete(
          Wrappers.<SysGroupUser>lambdaQuery()
              .eq(SysGroupUser::getType, GroupUserType.PRIMARY)
              .eq(SysGroupUser::getGroupId, groupId));
    }
    // 4. 生成新的
    AuthUser authUser = AuthUtil.getUser();
    LocalDateTime nowTime = LocalDateTime.now();
    Validate.isTrue(
        sysGroupUserMapper.insertBatchSomeColumn(
                ListUtil.map(
                    needAdUserIds,
                    userId -> {
                      SysGroupUser gu = new SysGroupUser();
                      gu.setUserId(userId);
                      gu.setGroupId(groupId);
                      gu.setCreateId(authUser.getIdAsLong());
                      gu.setCreateName(authUser.getFullName());
                      gu.setCreateTime(nowTime);
                      gu.setType(groupType);
                      gu.setTenantId(authUser.getTenantIdAsLong());
                      return gu;
                    }))
            > 0,
        "保存失败");
  }
  /**
   * 移除组中某个成员
   *
   * @param req {deptId,userId}
   */
  @Transactional(rollbackFor = Exception.class)
  public void userRemove(SysGroupUser req) {
    Validate.nonEmpty(req.getGroupId(), "用户组参数不能为空");
    Validate.nonEmpty(req.getUserId(), "用户参数不能为空");
    sysGroupUserMapper.delete(
        Wrappers.<SysGroupUser>lambdaQuery()
            .eq(SysGroupUser::getGroupId, req.getGroupId())
            .eq(SysGroupUser::getUserId, req.getUserId())
            .eq(SysGroupUser::getType, GroupUserType.NORMAL));
  }
  /**
   * 分页扩展
   *
   * @param page page
   * @param groupId groupId
   * @return IPage
   */
  public IPage<SysUserRes> userPage(Page<SysGroupUser> page, Long groupId) {
    return sysGroupUserMapper
        .selectPage(
            page,
            Wrappers.<SysGroupUser>lambdaQuery()
                .eq(SysGroupUser::getGroupId, groupId)
                .eq(SysGroupUser::getType, GroupUserType.NORMAL))
        .convert(
            item -> {
              SysUserRes res = new SysUserRes();
              var user =
                  Optional.ofNullable(sysUserMapper.selectById(item.getUserId()))
                      .orElse(new SysUser());
              res.setId(item.getUserId());
              res.setFullName(Objects.requireNonNullElse(user.getFullName(), "已注销"));
              res.setPhone(user.getPhone());
              res.setGroupIds(
                  sysGroupUserMapper
                      .selectList(
                          Wrappers.<SysGroupUser>lambdaQuery()
                              .eq(SysGroupUser::getUserId, item.getUserId())
                              .eq(SysGroupUser::getType, GroupUserType.NORMAL))
                      .stream()
                      .map(SysGroupUser::getGroupId)
                      .toList());
              // 获取关联的部门信息list
              Run.nonEmpty(
                  res.getGroupIds(),
                  (groupIds) ->
                      // 设置所有部门的名称并用逗号隔开
                      res.setGroupNames(
                          super.listByIds(groupIds).stream()
                              .map(SysGroup::getName)
                              .collect(Collectors.joining(Strings.COMMA))));
              return res;
            });
  }
  /**
   * 删除
   *
   * @param req req
   */
  public void del(SysGroup req) {
    Validate.nonNull(req.getId(), "id不能为空");
    Validate.isTrue(
        sysGroupUserMapper
                .lambdaQuery()
                .eq(SysGroupUser::getGroupId, req.getId())
                .eq(SysGroupUser::getType, DeptUserType.NORMAL)
                .count()
            <= 0,
        "删除失败，请先移除组内成员");
    Validate.isTrue(super.removeById(req.getId()), "删除失败，请重试");
  }
}
