package cn.zgfxrc.boot.upms.biz.service;

import cn.zgfxrc.boot.common.core.entity.base.BaseModel;
import cn.zgfxrc.boot.common.core.enums.PlatformType;
import cn.zgfxrc.boot.common.core.util.PageUtil;
import cn.zgfxrc.boot.common.core.util.TenantUtil;
import cn.zgfxrc.boot.upms.api.entity.SysTenant;
import cn.zgfxrc.boot.upms.api.entity.SysUser;
import cn.zgfxrc.boot.upms.api.entity.req.SysTenantReq;
import cn.zgfxrc.boot.upms.api.entity.req.SysUserReq;
import cn.zgfxrc.boot.upms.api.entity.res.SysTenantRes;
import cn.zgfxrc.boot.upms.biz.mapper.SysTenantMapper;
import cn.zgfxrc.boot.upms.biz.mapper.SysUserMapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.seepine.tool.Builder;
import com.seepine.tool.Run;
import com.seepine.tool.exception.RunException;
import com.seepine.tool.util.ListUtil;
import com.seepine.tool.util.Objects;
import com.seepine.tool.util.Validate;
import java.io.Serializable;
import java.time.LocalDate;
import java.util.Collections;
import lombok.AllArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author huanghs@zgfxrc.cn
 * @date 2021.11.2
 */
@Slf4j
@Service
@AllArgsConstructor
public class SysTenantService extends ServiceImpl<SysTenantMapper, SysTenant> {
  private final SysUserService sysUserService;
  private final SysRoleService sysRoleService;
  private final SysUserMapper sysUserMapper;
  private final SysDictService sysDictService;
  /**
   * 更新租户信息，并可能修改超管密码
   *
   * @param entity entity
   */
  @Transactional(rollbackFor = Exception.class)
  public void updateById(SysTenantReq entity) {
    Validate.nonEmpty(entity.getId(), "主键为空，请联系管理员");
    entity.setPackageIds(
        Validate.nonEmpty(
            ListUtil.filter(ListUtil.fromArray(entity.getPackageIds()), Objects::nonNull)
                .toArray(Long[]::new),
            "套餐不能为空"));
    Run.isEmpty(
        entity.getExpirationDate(), () -> entity.setExpirationDate(LocalDate.of(2999, 12, 31)));
    // 若有password，则修改密码
    sysUserService.updateAdminPassword(entity.getId(), entity.getPassword());

    SysTenant sysTenant = new SysTenant();
    sysTenant.setId(entity.getId());
    sysTenant.setPackageIds(entity.getPackageIds());
    sysTenant.setIsLock(entity.getIsLock());
    sysTenant.setLogo(entity.getLogo());
    sysTenant.setExpirationDate(entity.getExpirationDate());
    sysTenant.setActiveUserNumberLimit(entity.getActiveUserNumberLimit());
    Validate.isTrue(super.updateById(sysTenant), "保存失败");
  }

  /**
   * 删除租户
   *
   * @param id 租户id
   * @return bool
   */
  @Override
  public boolean removeById(Serializable id) {
    // 删除相关用户、角色、用户角色关联、角色菜单关联？不删除的话后续可以直接恢复数据
    return super.removeById(id);
  }

  /**
   * 保存租户，并初始化租户超管信息、角色、菜单等
   *
   * <p>仅供feign内部调用，不对外暴露接口
   *
   * @param entity entity
   * @date 2022.12.13 huanghs 增加过期时间及超管姓名
   */
  @SneakyThrows
  @Transactional(rollbackFor = Exception.class)
  public SysTenantRes add(SysTenantReq entity, PlatformType platformType) {
    entity.setShortName(Validate.nonBlank(entity.getName(), "名称不能为空").trim());
    entity.setIsSystem(false);
    entity.setIsLock(Objects.require(entity.getIsLock(), false));
    entity.setPlatformType(Validate.nonNull(platformType, "平台类型不能为空"));
    entity.setParentId(
        Objects.require(entity.getParentId(), Long.valueOf(TenantUtil.getTenantId())));
    entity.setExpirationDate(
        Objects.require(entity.getExpirationDate(), LocalDate.of(2999, 12, 31)));
    entity.setActiveUserNumberLimit(Objects.require(entity.getActiveUserNumberLimit(), 100));
    // 创建服务商
    Validate.isTrue(super.save(entity), "创建失败");
    TenantUtil.run(
        entity.getId(),
        () -> {
          // 1.创建超管用户
          try {
            sysUserService.add(
                Builder.of(SysUserReq::new)
                    .with(SysUserReq::setPhone, entity.getPhone())
                    .with(SysUserReq::setPassword, entity.getPassword())
                    .with(SysUserReq::setIsSystem, true)
                    .with(SysUserReq::setFullName, entity.getFullName())
                    .with(SysUserReq::setPlatformType, entity.getPlatformType())
                    .with(SysUserReq::setDeptIds, Collections.singletonList(0L))
                    .build());
          } catch (RunException e) {
            throw new RunException("创建超管账号失败:" + e.getMessage());
          }
          // 2.创建默认角色
          sysRoleService.createRole("行政", entity.getPlatformType());
          sysRoleService.createRole("人事", entity.getPlatformType());
          sysRoleService.createRole("财务", entity.getPlatformType());
          // 3.创建默认字典
          sysDictService.initDefaultDict(entity.getId());
        });
    SysTenantRes res = new SysTenantRes();
    BeanUtils.copyProperties(entity, res);
    return res;
  }

  public IPage<SysTenantRes> getPageRes(Page<SysTenant> page, SysTenant entity) {
    return PageUtil.convertPage(
        page(
            page,
            Wrappers.lambdaQuery(entity).eq(SysTenant::getPlatformType, PlatformType.ENTERPRISE)),
        (item) -> {
          SysTenantRes res = new SysTenantRes();
          BeanUtils.copyProperties(item, res);
          TenantUtil.run(
              item.getId(),
              () ->
                  Run.nonEmpty(
                      sysUserMapper
                          .lambdaQuery()
                          .eq(SysUser::getPlatformType, PlatformType.ENTERPRISE)
                          .eq(SysUser::getIsSystem, true)
                          .one(),
                      (user) -> {
                        res.setFullName(user.getFullName());
                        res.setPhone(user.getPhone());
                      }));
          return res;
        });
  }
  /**
   * 更新租户信息，场景例如服务商修改商户信息或渠道商信息，或商户修改自己企业信息
   *
   * @param entity entity
   */
  @Transactional(rollbackFor = Exception.class)
  public void updateTenant(SysTenantReq entity) {
    Validate.nonNull(entity.getId(), "参数错误，主键不能为空");
    Validate.nonNull(entity.getPlatformType(), "参数错误，平台类型不能为空");
    // 若有password，则修改密码
    sysUserService.updateAdminPassword(entity.getId(), entity.getPassword());
    var wrapper =
        super.lambdaUpdate()
            .eq(BaseModel::getId, entity.getId())
            .eq(SysTenant::getPlatformType, entity.getPlatformType())
            // 避免什么信息都没传导致sql错误
            .set(SysTenant::getPlatformType, entity.getPlatformType());
    Run.nonBlankAndTrim(entity.getName(), val -> wrapper.set(SysTenant::getName, val));
    Run.nonBlankAndTrim(entity.getShortName(), val -> wrapper.set(SysTenant::getShortName, val));
    Run.nonBlankAndTrim(entity.getLogo(), val -> wrapper.set(SysTenant::getLogo, val));
    Run.nonNull(entity.getIsLock(), val -> wrapper.set(SysTenant::getIsLock, val));
    Validate.isTrue(wrapper.update(), "保存失败");
  }
  /**
   * 修改租户名称
   *
   * @param id id
   * @param name 名称
   * @param shortName 简称
   */
  public void updateName(Long id, String name, String shortName) {
    Validate.nonEmpty(id, "信息错误");
    Validate.isTrue(
        lambdaUpdate()
            .eq(BaseModel::getId, id)
            .set(SysTenant::getName, Validate.nonBlank(name, "名称不能为空"))
            .set(SysTenant::getShortName, Validate.nonBlank(shortName, "简称不能为空"))
            .update(),
        "修改失败");
  }
}
